from pathlib import Path

from pydantic import Field, ValidationError
from pydantic_settings import BaseSettings

from . import utils

BASE = Path.home() / ".sous-chef"
GLOBAL_SETTINGS_FILE = BASE / "settings.json"


class Settings(BaseSettings):
    recipe_library: Path = Field(BASE / "recipes")
    plans_dir: Path = Field(BASE / "plans")

    def save(self) -> None:
        utils.touch(GLOBAL_SETTINGS_FILE)
        with open(GLOBAL_SETTINGS_FILE, "w") as file:
            file.write(self.model_dump_json())

    @classmethod
    def from_file(cls) -> "Settings":
        try:
            with open(GLOBAL_SETTINGS_FILE) as file:
                json_str = file.read()
                return cls.model_validate_json(json_str)

        except (FileNotFoundError, ValidationError):
            return cls()  # defaults



import yaml

from .recipe import Recipe
from .utils import capitalise


def serialize_yaml(recipe: Recipe) -> str:
    recipe_dict = recipe.model_dump()
    recipe_dict = {key: value for key, value in recipe_dict.items() if value}
    # make sure the ingredients are displayed in their yaml format
    recipe_dict["ingredients"] = [
        ing.to_yaml_str() for ing in recipe.ingredients
    ]
    return yaml.safe_dump(recipe_dict)


def serialize_json(recipe: Recipe) -> str:
    return recipe.model_dump_json()


def serialize_markdown(recipe: Recipe) -> str:
    strings = []

    title_block = f"# {capitalise(recipe.name)}"
    if recipe.image:
        title_block += f"\n![](../images/{recipe.image})"
    title_block += f"\n\nAuthor: {recipe.author.title()}"
    strings.append(title_block)

    if recipe.source:
        strings.append(f"From: {recipe.source}")
    if recipe.servings:
        strings.append(f"Servings: {recipe.servings}")
    if recipe.prep_minutes:
        strings.append(f"Preparation: {recipe.prep_minutes} minutes")
    if recipe.cook_minutes:
        strings.append(f"Cooking: {recipe.cook_minutes} minutes")
    if recipe.notes:
        strings.append(f"Notes: {recipe.notes}")

    if recipe.equipment:
        equipment_str = "## Equipment: "
        for item in recipe.equipment:
            equipment_str += f"\n- {capitalise(item)}"
        strings.append(equipment_str)

    ingredients_str = "## Ingredients:"
    for ing in recipe.ingredients:
        ingredients_str += f"\n- [ ] {capitalise(str(ing))}"
    strings.append(ingredients_str)

    if recipe.method:
        method_str = "## Method:"
        for ii, step in enumerate(recipe.method, start=1):
            method_str += f"\n{ii}. {capitalise(step)}"
        strings.append(method_str)

    return "\n\n".join(strings)



import asyncio
from pathlib import Path
from typing import Optional

import typer

from . import new, routines
from . import view
from .. import api
from ..api.settings import Settings

app = typer.Typer()
app.add_typer(new.app, name="new", help="Create new plan, recipe, etc")
app.add_typer(view.app, name="view", help="View plan, shopping list, etc")


@app.command()
def init(
    path: Path = typer.Argument(
        help="The path in which to initialise a new recipe library"
    ),
):
    """
    Initialise a new recipe library
    """
    path = path.absolute()
    print(f"you passed {path=}")
    api.init_library(path)


@app.command()
def config(
    recipe_library: str = typer.Option(
        "",
        "--recipe-library",
        help="The folder in which the YAML recipes are stored",
    ),
):
    """
    Update and/or view configuration
    """
    settings = Settings.from_file()
    if recipe_library:
        print(f"{recipe_library=}")
        if (recipe_library := Path(recipe_library)).exists():
            absolute_path = recipe_library.absolute()
            settings.recipe_library = absolute_path
            print(f"Saving recipe library: {absolute_path}")
            settings.save()
        else:
            typer.echo(f"{recipe_library} does not exist")

    print("config: ")
    for key, val in settings.model_dump().items():
        print(f"\t{key}: {val}".expandtabs(4))


@app.command()
def plan(
    query: list[str] = typer.Argument(help="Search term to find recipes"),
):
    """
    Add a recipe to a plan
    """
    query = " ".join(query)
    asyncio.run(routines.plan_recipe(query))


@app.command()
def export():
    """
    Convert YAML recipes to markdown
    """
    asyncio.run(routines.export_recipes())


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{api.__app_name__} v{api.__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the application's version and exit.",
        callback=_version_callback,
        is_eager=True,
    ),
) -> None:
    return



import string
from pathlib import Path


def touch(file: Path):
    file.parent.mkdir(exist_ok=True, parents=True)
    file.touch(exist_ok=True)


def clean_filename(original: Path) -> Path:
    """
    Strip spaces and punctuation out of the filename and replace them with
    hyphens.

    :param original: E.g. "some/folder/my bad file.txt"
    :return: the cleaned filename e.g. "some/folder/my-bad-file.txt"
    """
    mapping = {
        " ": "-",
        "_": "-",
        ".": "-",
    }
    stem = original.stem.lower()
    allowed = string.ascii_letters + string.digits + "-"
    cleaned = "".join(
        char if char in allowed else mapping.get(char, "") for char in stem
    )
    out = original.parent / f"{cleaned}{original.suffix}"
    return out



from pydantic import BaseModel, Field

from .recipe import Recipe

Amount = int | float
Unit = str
IngredientName = str
RecipeName = str


class Amounts(BaseModel):
    # from "apples" -> display as "enough for x"
    enough_for: list[RecipeName] = []

    # from "4 apples"
    unitless: Amount = 0

    # from "4 kg apples"
    units: dict[Unit, Amount] = Field(default_factory=dict)


ShoppingList = dict[IngredientName, Amounts]


example = ShoppingList(
    apples=Amounts(
        enough_for=["apple pie"],
        unitless=4,
        units=dict(kg=4),
    ),
    potatoes=Amounts(units=dict(kg=5)),
)


def merge_recipes(recipes: list[Recipe]) -> ShoppingList:
    """
    This is the magic.
    :param recipes:
    :return:
    """
    shopping_list = {}
    for recipe in recipes:
        for ing in recipe.ingredients:
            amounts = shopping_list.get(ing.name, Amounts())
            match [ing.amount, ing.unit]:
                case [0, ""]:
                    amounts.enough_for.append(recipe.name)
                case [amount, ""]:
                    amounts.unitless += amount
                case [amount, unit]:
                    amounts.units[unit] = amounts.units.get(unit, 0) + amount
            shopping_list[ing.name] = amounts

    return shopping_list



"""
view.py
"""

import typer

from .. import api
from ..api.shopping_list import Amounts
from .utils import echo

app = typer.Typer()


@app.command(name="plan")
def view_plan():
    """
    view_plan
    """
    plan = api.Plan.current()
    echo("Current plan:")
    echo(f"\tcreated: {plan.created.isoformat()}")
    if plan.recipes:
        echo("\trecipes:")
        for recipe in plan.recipes:
            echo(f"\t\t{recipe}")


@app.command(name="list")
def view_list():
    """
    view_list
    """
    plan = api.Plan.current()
    ingredients = plan.shopping_list()
    echo("Current shopping list:")
    width = max(map(len, ingredients))
    for ing_name in sorted(ingredients):
        amounts = ingredients[ing_name]
        echo(_format_ingredient_for_list(ing_name, amounts, width))


@app.command(name="recipe")
def view_recipe():
    """
    view recipe
    """
    # todo: serialize recipe nicely for terminal


def _format_ingredient_for_list(
    ing_name: str,
    amounts: Amounts,
    width: int = 0,
) -> str:
    result = f"{ing_name}:".ljust(width)
    unique_recipes = len(set(amounts.enough_for))
    match [unique_recipes, amounts.unitless, len(amounts.units)]:
        case [0, 0, 0]:
            raise ValueError("empty Amounts!")

        # ========================= single line cases =========================
        # only 1 unique recipe (could be repeated)
        case [1, 0, 0]:
            recipe_strings = (
                amounts.enough_for[0]
                if (count := len(amounts.enough_for)) == 1
                else f"{amounts.enough_for[0]} (x{count})"
            )
            result += f" enough for {recipe_strings}"

        # just the unitless amount
        case [0, unitless, 0]:
            result += f" {unitless}"

        # only 1 unit e.g. {"kg": 2.5}
        case [0, 0, 1]:
            unit, amount = next(iter(amounts.units.items()))
            result += f" {amount} {unit}"

        # ========================== multi line cases ==========================
        case _:
            if amounts.enough_for:
                result += f"\n\t{_format_amountless(amounts.enough_for)}"
            if amounts.unitless:
                result += f"\n\t{amounts.unitless}"
            for unit, amount in amounts.units.items():
                result += f"\n\t{amount} {unit}"

    return result


def _format_amountless(names: list[str]) -> str:
    def _format(name: str, count: int) -> str:
        return name if count == 1 else f"{name} (x{count})"

    counts = _count(names)
    if len(counts) == 1:
        name, count = next(iter(counts.items()))
        return f"enough for {_format(name, count)}"
    else:
        result = "enough for:"
        for name, count in counts.items():
            s = _format(name, count)
            result += f"\n\t\t{s}"
        return result


def _count(names: list[str]) -> dict[str, int]:
    counts = {}
    for name in names:
        counts[name] = counts.get(name, 0) + 1
    return counts



from pathlib import Path


def init_library(path: Path):
    for p in [
        path,
        path / "yaml",
        path / "md",
        path / ".chef",
    ]:
        p.mkdir(parents=True, exist_ok=True)  # ensure the dir exists



import asyncio
from pathlib import Path

import aiofiles
import yaml
from pydantic import BaseModel, Field

from .utils import preprocess_yaml
from ..ingredient import Ingredient
from ..settings import Settings


class Recipe(BaseModel):
    # metadata
    name: str
    author: str = ""
    prep_minutes: int = 0
    cook_minutes: int = 0
    servings: int = 0
    source: str = ""
    image: str = ""
    notes: str = ""

    # child objects
    ingredients: list[Ingredient] = Field()

    # for something like a salsa or a spice mix, the "method" is just:
    # combine all the ingredients. So method should not be mandatory.
    method: list[str] = []
    equipment: list[str] = []

    def __str__(self):
        return f"{self.name} by {self.author}"

    @staticmethod
    def search(query: str, recipes: list["Recipe"]) -> list["Recipe"]:
        """
        :param query: Could be anything -- ingredient, title fragment, author
        :param recipes: the recipes to search
        :return: list of recipes that match the given query term
        """
        # todo: improve this function. Should be able to search specifically for
        #  author, or ingredient. Should be able to pass via flags e.g.
        #       -a --author Robin Neville
        #       -i --ingredient lentils
        #       -e --equipment air fryer
        #       -x --exclude (flips the above logic as if applying `not` to the
        #                     whole query)

        if not query or not recipes:
            return []

        return [
            recipe
            for recipe in recipes
            if any(
                needle.lower() in haystack.lower()
                for needle in query.split()
                for haystack in [recipe.name, recipe.author]
            )
        ]

    @classmethod
    async def load_all(cls) -> list["Recipe"]:
        """
        Load all the recipes from yaml
        """
        settings = Settings.from_file()
        # todo: de-dupe these paths
        filenames = settings.recipe_library.joinpath("yaml").glob("*.yaml")
        return list(await asyncio.gather(*(cls.load(f) for f in filenames)))

    @classmethod
    async def load(cls, filename: Path) -> "Recipe":
        """
        Load from yaml
        """
        async with aiofiles.open(filename) as file:
            recipe_str = await file.read()
            recipe_dict = yaml.safe_load(recipe_str)
            recipe_dict = preprocess_yaml(recipe_dict)
            return Recipe(**recipe_dict)



from pydantic import BaseModel


class Ingredient(BaseModel):
    name: str
    amount: int | float = 0
    unit: str = ""
    prep: str = ""

    @classmethod
    def from_str(cls, s: str) -> "Ingredient":
        """
        :raises: ParseIngredientError
        """
        return Ingredient(**parse_ingredient_str(s))

    def __str__(self):
        match [self.name, self.amount, self.unit]:
            case [name, 0, ""]:
                result = f"{name}"
            case [name, amount, ""]:
                result = f"{amount} {name}"
            case [name, amount, unit]:
                result = f"{amount} {unit} {name}"
            case _:
                raise RuntimeError("unreachable")

        if self.prep:
            result += f", {self.prep}"

        return result

    def to_yaml_str(self):
        match [self.name, self.amount, self.unit]:
            case [name, 0, ""]:
                result = f"{name}"
            case [name, amount, ""]:
                result = f"{amount}, {name}"
            case [name, amount, unit]:
                result = f"{amount}, {unit}, {name}"
            case _:
                raise RuntimeError("unreachable")

        if self.prep:
            result += f"; {self.prep}"

        return result


def parse_ingredient_str(s: str) -> dict:
    """
    :param s: examples:
                name only:                    "apple"
                amount and name:              "1, apple"
                amount, unit and name:        "1, kg, apples"
                amount, unit, name, and prep: "1, kg, apples; chopped"
    :return:
    :raises: IngredientParseError
    """
    output = {}

    match list(map(str.strip, s.split(";"))):
        case [""]:
            raise ParseIngredientError("empty input")
        case [s]:
            pass
        case [s, ""]:  # trailing ";" and empty prep
            pass
        case [s, prep]:  # non-empty prep
            output["prep"] = prep
        case _:
            raise ParseIngredientError(f"Too many semicolons: {s}")

    match list(map(str.strip, s.split(","))):
        case [""]:
            raise ParseIngredientError("Got prep but no ingredients")
        case [amount, name]:  # e.g. 1, apple
            if not name:
                raise ParseIngredientError("empty value: name")
            if not amount:
                raise ParseIngredientError("empty value: amount")
            output["name"] = name
            output["amount"] = _parse_number(amount)
        case [amount, unit, name]:  # e,g. 1, kg, apples
            output["name"] = name
            output["amount"] = _parse_number(amount)
            output["unit"] = unit
        case [name]:
            output["name"] = name
        case _:
            raise ParseIngredientError(f"Too many commas: {s}")

    return output


class ParseIngredientError(Exception):
    pass


def _parse_number(number: str) -> int | float:
    """
    :raises: IngredientParseError if invalid number
    """
    if not number.strip():
        return 0

    try:
        result = float(number)
    except ValueError:
        raise ParseIngredientError(f"invalid number: {number!r}")
    else:
        if result % 1 == 0:
            return int(result)
        else:
            return result



import typer
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

from ...api import Plan
from ...api.recipe import Recipe
from ..utils import multiple_choice_menu


async def plan_recipe(query: str):
    plan = Plan.current()

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=True,  # will disappear on complete
    ) as progress:
        progress.add_task("loading recipes")
        all_recipes = await Recipe.load_all()

    matches = Recipe.search(query=query, recipes=all_recipes)

    match len(matches):
        case 0:
            typer.secho(f"Couldn't find any recipes for {query}")
            return
        case 1:
            # if there's only one match, don't bother prompting the user
            recipe = matches[0]
        case _:
            # todo: allow Ctrl+C and abort gracefullys
            recipe = multiple_choice_menu(
                prompt="Which recipe did you mean?",
                choices={r.name: r for r in matches},
            )

    plan.add(recipe)
    plan.save()
    typer.secho(f"Added {recipe} to plan. Plan is now: ")
    typer.secho(f"{plan}")



from pathlib import Path
from typing import Any

from src.api import Recipe, RecipeSerializer, Settings
from src.api.ingredient import ParseIngredientError, Ingredient
from src.api.recipe.serializer import Format
from src.api.utils import clean_filename


def recipe_wizard() -> Recipe:
    """
    Interactively guide the user through creating a new recipe
    """
    name = prompt("name")
    author = prompt("author")

    print("Ingredients:")
    print("format: [amount, [unit,]] name [; prep]")
    ingredients = []
    while ing_str := prompt("enter an ingredient [leave blank to finish]"):
        try:
            ingredient = Ingredient.from_str(ing_str)
        except ParseIngredientError as e:
            print(e)
            continue
        else:
            ingredients.append(ingredient)

    print("Method:")
    method = []
    while method_str := prompt("enter a method step [leave blank to finish]"):
        method.append(method_str)

    print("Equipment:")
    equipment = []
    while equipment_str := prompt(
        "what equipment do you need? [leave blank to finish]"
    ):
        equipment.append(equipment_str)

    recipe = Recipe(
        name=name,
        author=author,
        ingredients=ingredients,
        method=method,
        equipment=equipment,
        prep_minutes=prompt("prep_minutes", default=0, expected_type=int),
        cook_minutes=prompt("cook_minutes", default=0, expected_type=int),
        servings=prompt("servings", default=0, expected_type=int),
        source=prompt("source (url or book name)"),
        image=prompt("image (url or local image)"),
        notes=prompt("notes"),
    )
    return recipe


NOT_SET = object()


def prompt(name: str, default: Any = NOT_SET, expected_type: type = str) -> Any:
    s = f"{name}"
    if expected_type != str:
        s += f" ({expected_type.__name__})"
    if default is not NOT_SET:
        s += f" [{default}]"
    s += ": "

    while True:
        raw = input(s)
        if not raw:
            if default is NOT_SET:
                return expected_type()
            else:
                return default

        try:
            parsed = expected_type(raw)
        except Exception as e:
            print(e)
        else:
            return parsed



import json
from datetime import datetime

from pydantic import BaseModel, Field

from . import utils
from .recipe import Recipe
from .settings import Settings
from .shopping_list import merge_recipes, ShoppingList


class Plan(BaseModel):
    created: datetime = Field(default_factory=lambda: datetime.utcnow())
    recipes: list[Recipe] = Field(default_factory=list)

    def __str__(self):
        return f"Plan (created at {self.created})"

    def shopping_list(self) -> ShoppingList:
        """
        Merge the ingredients from self.recipes into one list of ingredients.
        """
        return merge_recipes(self.recipes)

    def add(self, recipe: Recipe):
        self.recipes.append(recipe)

    @classmethod
    def new(cls) -> "Plan":
        """
        Create a new plan in the filesystem
        :return: new plan
        """
        plan = cls()
        filename = plan.filename
        utils.touch(filename)
        with open(filename, "w") as file:
            file.write(plan.model_dump_json())
        return plan

    @classmethod
    def current(cls) -> "Plan":
        settings = Settings.from_file()
        json_files = settings.plans_dir.glob("*.json")
        latest = max(json_files)  # relying on string comparison here
        with open(latest) as file:
            try:
                return Plan(**json.load(file))
            except json.JSONDecodeError:
                return cls.new_plan()

    def save(self):
        with open(self.filename, "w") as file:
            file.write(self.model_dump_json())

    @property
    def filename(self):
        settings = Settings.from_file()
        return settings.plans_dir / f"{self.created.isoformat()}.json"
